% NLM_SURELET_Fixed
% --------------------------------------------------------------------------------
% If you use this software, please cite:
%
% D. Van De Ville, M. Kocher, 
% Non-Local Means with Dimensionality Reduction and SURE-Based Parameter Selection 
% IEEE Transactions on Image Processing, in press
% --------------------------------------------------------------------------------
%
% Run (multiple) NLM algorithms and (linearly) combine their outputs based
% on Stein's Unbiased Risk Estimate (SURE). 
%
% Interactive arguments:
%
% - TreeParameter: 
%   Selects the basic parameters about the experiment from TreeParam.m
%   The "TreeParameters" defines the noise level of the additive Gaussian
%   noise; the dimensionality of the PCA subspace projection (if used); 
%   and the test image to be used.
%
% - LeafParameter: 
%   Selects the parameters of the NLM algorithms from LeafParam.m
%   Each row represents a NLM algorithms with three required parameters:
%    B which characterizes the size of the patch (size=(2B+1)x(2B+1))
%    S which characterizes the size of the search (size=(2S+1)x(2S+1))
%    SigmaFactor characterizes the strength of the smoothing as a function
%                of the noiselevel
%   The optional fourth argument allows to specify the dimensionality of
%   the PCA subspace for each NLM independently. If the parameters are set
%   to zero, then the identity algorithm is used (output=input).
%
% --------------------------------------------------------------------------------


%% Init
clear
close all
path (path, 'toolbox/');
Fig = 1;

%% Parameters
Display = 1;

TreeParameterName = input ('Tree parameter name (e.g., 1) = ');
P = TreeParam (TreeParameterName);

LeafParameterName = input ('Leaf parameter name (e.g., 1) = ');
ParameterSet = LeafParam (LeafParameterName);

%% Directory
% DirectoryName = ['Result/Result',num2str(TreeParameterName),num2str(LeafParameterName)];
% if exist (DirectoryName) == 0
%     mkdir (DirectoryName);
% elseif exist (DirectoryName) == 7
%     rmdir (DirectoryName, 's');
%     mkdir (DirectoryName);
% end;

%% Data
f = double (imread (['Images/',P.Name{1},'.png']));
if P.SigmaNoise > 0
    randn ('seed', 0);
    noise = P.SigmaNoise * randn (size (f));
    noise = P.SigmaNoise * noise / std(noise(:));
    fn = f + noise;
    TrueSigmaNoise = std (noise (:)); % true SigmaNoise
end

MSEOrig = mean (abs (fn(:) - f(:)).^2);
PSNROrig = -10 * log10 (MSEOrig / 255^2);

%% Process
tic
[PSNR, SUREPSNR, Theta, FkStack, Fky] = surelet_nlmeans2 (f, fn, P.SigmaNoise, ParameterSet, P);
toc
%% Display
[PSNR, SUREPSNR]

figure (Fig), Fig = Fig + 1;
montage (reshape (uint8(FkStack(:, :, :)), size(FkStack(:, :, :), 1), size (FkStack(:, :, :), 2), 1, size (FkStack(:, :, :), 3)))
% print (['-f',num2str(Fig-1)], '-depsc', [DirectoryName,'/Photo',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM),'.eps'])

figure (Fig), Fig = Fig + 1;
subplot (2, 2, 1)
bar (Theta), title ('\theta'), axis square, axis tight
subplot (2, 2, 2)
imshow (f, [0, 255]), title ('f')
subplot (2, 2, 3)
imshow (fn, [0, 255]), title ('f_n')
subplot (2, 2, 4)
imshow (Fky, [0, 255]), title ('LE')
% print (['-f',num2str(Fig-1)], '-depsc', [DirectoryName,'/Photo',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM),'Bis.eps'])
