% NLM_SURELET_MonteCarlo
% --------------------------------------------------------------------------------
% If you use this software, please cite:
%
% D. Van De Ville, M. Kocher, 
% Non-Local Means with Dimensionality Reduction and SURE-Based Parameter Selection 
% IEEE Transactions on Image Processing, in press
% --------------------------------------------------------------------------------
%
% Run (multiple) NLM algorithms with Monte-Carlo generated parameters and 
% (linearly) combine their outputs based on Stein's Unbiased Risk 
% Estimate (SURE). 
%
% Interactive arguments:
%
% - TreeParameter: 
%   Selects the basic parameters about the experiment from TreeParam.m
%   The "TreeParameters" defines the noise level of the additive Gaussian
%   noise; the dimensionality of the PCA subspace projection (if used); 
%   and the test image to be used.
%
% - LeafParameter: 
%   Selects the parameters of the NLM algorithms...
%   Each row represents a NLM algorithms with three required parameters:
%    B which characterizes the size of the patch (size=(2B+1)x(2B+1))
%    S which characterizes the size of the search (size=(2S+1)x(2S+1))
%    SigmaFactor characterizes the strength of the smoothing as a function
%                of the noiselevel
%   The optional fourth argument allows to specify the dimensionality of
%   the PCA subspace for each NLM independently. If the parameters are set
%   to zero, then the identity algorithm is used (output=input).
%   These parameters are generated randomly, see lines 80-91
%
% --------------------------------------------------------------------------------

%% Init
clear
close all
path (path, 'toolbox/');
Fig = 1;
NbTrials = 1; % 10;

%% Parameters
TheTreeParameterName = [1];
TheNbNLM = [3]; % [12];

%% Process
for cpt = 1 : numel (TheTreeParameterName)
    TreeParameterName = TheTreeParameterName(cpt);
    P = TreeParam (TreeParameterName);
    % Data
    f = double (imread (['Images/',P.Name{1},'.png']));
    [NR, NC] = size (f);
    if P.SigmaNoise > 0
        randn ('seed', 0);
        noise = P.SigmaNoise * randn (size (f));
        noise = P.SigmaNoise * noise / std(noise(:));
        fn = f + noise;
        TrueSigmaNoise = std (noise (:)); % true SigmaNoise
    end
    
    MSEOrig = mean (abs (fn(:) - f(:)).^2);
    PSNROrig = -10 * log10 (MSEOrig / 255^2);
    
    for cpt1 = 1 : numel (TheNbNLM)
        NbNLM = TheNbNLM(cpt1);
        DirectoryName = ['Result/Genetic',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM)];
        if exist (DirectoryName) == 0
            mkdir (DirectoryName);
        elseif exist (DirectoryName) == 7
            rmdir (DirectoryName, 's');
            mkdir (DirectoryName);
        end;
        
        % surelet_nlmeans2
        clear PSNR SUREPSNR Theta FkStack Fky psnr_res sure_psnr Cond
        tic
        for cpt2 = 1 : NbTrials
            [cpt, cpt1, cpt2]
            if NR == 256
                ParameterSet = zeros (NbNLM+1, 4);
                ParameterSet (2:end, 1) = unidrnd (3, NbNLM, 1);   % B between 1 and 3
                ParameterSet (2:end, 2) = unidrnd (9, NbNLM, 1)+1; % S between 2 and 10
                ParameterSet (2:end, 3) = 0.4*rand (NbNLM, 1)+.6;  % Lambda between .6 and 1
                
            elseif NR == 512
                ParameterSet = zeros (NbNLM+1, 4);
                ParameterSet (2:end, 1) = unidrnd (4, NbNLM, 1);   % between 1 and 6
                ParameterSet (2:end, 2) = unidrnd (15, NbNLM, 1)+1; % between 2 and 20
                ParameterSet (2:end, 3) = 0.4*rand (NbNLM, 1)+.6;  % between .6 and 1
            else
                error ('size not yet supported')
            end;
            
            Min = 3;
            for cpt3 = 2 : NbNLM+1
                Max = (ParameterSet (cpt3, 1)*2+1)^2; % (2*B+1)^2
                ParameterSet (cpt3, 4) = Min + unidrnd (Max-Min);  % between Min and Max
            end;
            
            [PSNR(cpt2), SUREPSNR(cpt2), Theta(:, cpt2), FkStack(:, :, :, cpt2), Fky(:, :, cpt2), psnr_res(:, cpt2), sure_psnr(:, cpt2), Cond(cpt2, :)] = ...
                surelet_nlmeans2 (f, fn, TrueSigmaNoise, ParameterSet, P);
        end;
        toc
        
        % Keep the best
        IxBest = find (PSNR == max (PSNR));
        IxBest = IxBest (1);
        
        % Save to file
        Save.FkStack      = FkStack (:, :, :, IxBest);
        Save.Fky          = Fky (:, :, IxBest);
        Save.Theta        = Theta (:, IxBest);
        Save.fn           = fn;
        Save.f            = f;
        Save.ParameterSet = ParameterSet;
        Save.PSNR         = PSNR;
        Save.SUREPSNR     = SUREPSNR;
        Save.psnr_res     = psnr_res;
        Save.sure_psnr    = sure_psnr;
        Save.Cond         = Cond;
        save ([DirectoryName,'/Data',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM),'.mat'], 'Save');
        
        % Display
        [PSNR, SUREPSNR]
        figure (Fig), Fig = Fig + 1;
        montage (reshape (uint8(FkStack(:, :, :, IxBest)), size(FkStack(:, :, :, IxBest), 1), size (FkStack(:, :, :, IxBest), 2), 1, size (FkStack(:, :, :, IxBest), 3)))
        print (['-f',num2str(Fig-1)], '-depsc', [DirectoryName,'/Photo',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM),'.eps'])
        
        figure (Fig), Fig = Fig + 1;
        subplot (2, 2, 1)
        plot (1:NbTrials, PSNR, 'r', 1:NbTrials, SUREPSNR, 'g'), title ('PSNR, SUREPSNR'), axis square, axis tight
        subplot (2, 2, 2)
        bar (Theta(:, IxBest)), title ('\theta'), axis square, axis tight
        subplot (2, 3, 4)
        imshow (f, [0, 255]), title ('f')
        subplot (2, 3, 5)
        imshow (fn, [0, 255]), title ('f_n')
        subplot (2, 3, 6)
        imshow (Fky (:, :, IxBest), [0, 255]), title ('LE')
        print (['-f',num2str(Fig-1)], '-depsc', [DirectoryName,'/Photo',num2str(TreeParameterName),'_',num2str(NbTrials),'_',num2str(NbNLM),'Bis.eps'])
    end;
end;